/*
 * Svjatoslav Commons - shared library of common functionality.
 * Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 */

package eu.svjatoslav.commons.string.tokenizer;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;
import java.util.stream.Stream;

import static eu.svjatoslav.commons.string.tokenizer.Terminator.TerminationStrategy.DROP;
import static eu.svjatoslav.commons.string.tokenizer.Terminator.TerminationStrategy.PRESERVE;

public class Tokenizer {

    private final Stack<Integer> tokenIndexes = new Stack<>();
    private final List<Terminator> terminators = new ArrayList<>();
    private String source;
    private int currentIndex = 0;

    private int cachedTerminatorIndex = -1;
    private Terminator cachedTerminator;

    public Tokenizer(final String source) {
        this.source = source;
    }

    public Tokenizer() {
    }

    public Tokenizer setSource(String source) {
        this.source = source;
        currentIndex = 0;
        tokenIndexes.clear();

        cachedTerminatorIndex = -1;
        cachedTerminator = null;
        return this;
    }

    public Tokenizer addTerminator(final String startSequence,
                                   final Terminator.TerminationStrategy terminationStrategy) {
        terminators.add(new Terminator(startSequence, terminationStrategy));
        return this;
    }

    public Tokenizer addTerminator(final String startSequence,
                                   final String endSequence, final Terminator.TerminationStrategy terminationStrategy) {
        terminators.add(new Terminator(startSequence, endSequence, terminationStrategy));
        return this;
    }

    public void expectAndConsumeNextToken(final String value)
            throws InvalidSyntaxException {
        final TokenizerMatch match = getNextToken();
        if (!value.equals(match.token))
            throw new InvalidSyntaxException("Expected \"" + value
                    + "\" but got \"" + match.token + "\" instead.");
    }


    public TokenizerMatch getNextToken() throws InvalidSyntaxException {
        tokenIndexes.push(currentIndex);

        StringBuilder tokenAccumulator = new StringBuilder();

        while (true) {

            if (currentIndex >= source.length()) { // reached end of input
                if (hasAccumulatedToken(tokenAccumulator))
                    return new TokenizerMatch(tokenAccumulator.toString(), null, null);
                else
                    return null;
            }

            if (isOngoingToken()) {
                tokenAccumulator.append(source.charAt(currentIndex));
                currentIndex++;
                continue;
            }

            Terminator terminator = getOrFindTokenTerminator();

            if (terminator.termination == PRESERVE)
                return buildPreservedToken(tokenAccumulator, terminator);
            else if (terminator.termination == DROP) {
                skipUntilTerminatorEnd(terminator);

                if (hasAccumulatedToken(tokenAccumulator))
                    return new TokenizerMatch(tokenAccumulator.toString(), null, terminator);
            }
        }

    }

    private void skipUntilTerminatorEnd(Terminator terminator) throws InvalidSyntaxException {
        if (terminator.hasEndSequence())
            currentIndex = getEndSequenceIndex(terminator) + terminator.endSequence.length();
        else
            currentIndex += terminator.startSequence.length();
    }

    private TokenizerMatch buildPreservedToken(StringBuilder token, Terminator terminator) throws InvalidSyntaxException {
        if (hasAccumulatedToken(token))
            return new TokenizerMatch(token.toString(), null, terminator);

        if (terminator.hasEndSequence())
            return buildComplexPreservedToken(terminator);
        else
            return buildSimplePreservedToken(terminator);
    }

    private TokenizerMatch buildSimplePreservedToken(Terminator terminator) {
        currentIndex += terminator.startSequence.length();
        return new TokenizerMatch(terminator.startSequence, null, terminator);
    }

    private TokenizerMatch buildComplexPreservedToken(Terminator terminator) throws InvalidSyntaxException {
        int endSequenceIndex = getEndSequenceIndex(terminator);
        String reminder = source.substring(currentIndex + terminator.startSequence.length(), endSequenceIndex);
        currentIndex = endSequenceIndex + terminator.endSequence.length();

        return new TokenizerMatch(terminator.startSequence, reminder, terminator);
    }

    private int getEndSequenceIndex(Terminator terminator) throws InvalidSyntaxException {
        int endSequenceIndex = source.indexOf(terminator.endSequence,
                currentIndex + terminator.startSequence.length());

        if (endSequenceIndex < 0)
            throw new InvalidSyntaxException("Expected \"" + terminator.endSequence + "\" but not found.");

        return endSequenceIndex;
    }

    private boolean hasAccumulatedToken(StringBuilder token) {
        return token.length() > 0;
    }

    private boolean isOngoingToken() {
        return getOrFindTokenTerminator() == null;
    }

    public boolean hasMoreTokens() {
        return currentIndex < source.length();
    }

    /**
     * Attempts to cache terminator search result.
     */
    public Terminator getOrFindTokenTerminator() {
        if (currentIndex == cachedTerminatorIndex)
            return cachedTerminator;

        cachedTerminatorIndex = currentIndex;
        cachedTerminator = findTokenTerminator();
        return cachedTerminator;
    }

    private Terminator findTokenTerminator() {
        for (Terminator terminator : terminators)
            if (terminator.matches(source, currentIndex))
                return terminator;
        return null;
    }

    public boolean consumeIfNextToken(final String token) throws InvalidSyntaxException {
        if (token.equals(getNextToken().token))
            return true;

        unreadToken();
        return false;
    }

    public TokenizerMatch peekNextToken() throws InvalidSyntaxException {
        TokenizerMatch result = getNextToken();
        unreadToken();
        return result;
    }

    public boolean peekIsOneOf(String... possibilities) throws InvalidSyntaxException {
        String nextToken = peekNextToken().token;
        return Stream.of(possibilities).anyMatch(possibility -> possibility.equals(nextToken));
    }

    public void peekExpectNoneOf(String... possibilities) throws InvalidSyntaxException {
        if (peekIsOneOf(possibilities))
            throw new InvalidSyntaxException("Not expected \"" + peekNextToken().token + "\" here.");
    }

    public void unreadToken() {
        currentIndex = tokenIndexes.pop();
    }

    public void skipUntilDataEnd() {
        tokenIndexes.push(currentIndex);
        currentIndex = source.length();
    }

}
